
-- TODO detect end of run by when player loses control (find ram location on 1-1 boss)
-- FIXME use savestate.registersave() and savestate.registerload() to make the script work across savestates (for testing memory conditions)

-- automatically save movies of runs
-- relative to the path of this script
-- set to nil to disable recording of movies
MoviePath = '../runs'
MinimumMovieLength = 5 -- minutes

MessageTime = 3 -- seconds

FramesPerSecond = 60.0988 -- NES
require('estimate-parameters')
BossTime = {BossTime1, BossTime2, BossTime3, BossTime4}

function main()
   if MoviePath then
      if MoviePath:sub(MoviePath:len()) ~= '/' then
         MoviePath = MoviePath..'/'
      end
      -- check if directory exists
      local success, err, code = os.rename(MoviePath, MoviePath)
      if not success and code ~= 13 then
         showMessage('MoviePath doesn\'t exist: '..MoviePath)
         MoviePath = nil
      end
   end

   gui.register(draw)

   resetGame()
   while true do
      emu.frameadvance()
      detectReset()
      detectStart()
      detectSplit()
      detectCutscene()
      updateMessages()
   end
end

function draw()
   if started then
      status = levelToString(level)
         ..' '..formatTime(emu.framecount() - started)
         ..' '..formatTime(estimateTime())
         ..' '
   else
      status = '0-0  0:00  0:00 '
   end
   -- letters draw from their lower right corner, and are 8x8
   gui.text(164, 224, status)

   for i, msg in ipairs(messages) do
      gui.text(8, i*8, msg.text)
   end
end

messages = {}
function showMessage(text)
   table.insert(messages, {text=text, frame=emu.framecount()})
end

function updateMessages()
   for i = #messages, 1, -1 do
      if emu.framecount() - messages[i].frame >= FramesPerSecond * MessageTime then
         table.remove(messages, i)
      end
   end
end

function resetGame()
   -- detectStart() sets and others use
   started = nil -- frame number or nil
   startstate = 'title'
   -- detectSplit() sets and others use
   level = 0
   times = {}
   -- detectCutscene() sets and others use
   cutsceneTimer = nil

   -- only fires if a new game was started (not on power cycles)
   if movie.mode() == 'record' then
      movie.stop()
   end
   
   if movieFilename and os.time() - movieStarted < MinimumMovieLength * 60 then
      showMessage('Removing short movie.')
      os.remove(movieFilename)
   end
   movieFilename = nil
   
   if MoviePath and movie.mode() ~= 'playback' then
      local now = os.date('*t')
      movieFilename = MoviePath..string.format('%d-%02d-%02d %02d-%02d-%02d.fm2', now.year, now.month, now.day, now.hour, now.min, now.sec)
      movie.record(movieFilename)
      movieStarted = os.time()
      showMessage('Recording '..movieFilename)
   end
end

function detectReset()
   -- this detects: console reset, new game after gameover
   -- FIXME if you reset before taking control on 1-1, this fails to reset (started isn't set)
   -- FIXME if you gameover on 1-1 and start a new game, this fails to reset (it's not a hard reset, and level is still 0)
   if started and (emu.framecount() < started or getCurrentLevel() < level) then
      resetGame()
   end
end

function detectStart()
   -- FIXME the 5th demo on the title screen triggers this start-detection
   if not started then
      if startstate == 'title' and getFlags() == '0,0' then
         startstate = 'waiting'
      elseif startstate == 'waiting' and getFlags() == '2,0' then
         startstate = 'play'
         started = emu.framecount()
      end
   end
end

function detectSplit()
   if started and getCurrentLevel() ~= level then
      table.insert(times, {name=levelToString(level), time=(emu.framecount() - started)})
      level = getCurrentLevel()
   end
end

function detectCutscene()
   if isCutsceneActive() then
      if not cutsceneTimer then
         cutsceneTimer = 0
      end
      cutsceneTimer = cutsceneTimer + 1
   else
      cutsceneTimer = nil
   end
end


-- return frame count at estimated run completion time
function estimateTime()
   local baseTime = 0  -- seconds
   if started then
      baseTime = (emu.framecount() - started) / FramesPerSecond
   end

   for num = level, 15 do
      baseTime = baseTime + estimateLevelTime(num)
   end

   return baseTime * FramesPerSecond
end

function estimateLevelTime(num)
   local remaining = 1
   if num == level then remaining = getLevelRemaining() end

   local t = remaining * LevelTime
   
   if remaining > 0 then
      t = t + BossTime[1 + num % 4]
   end
   
   if num ~= 15 then
      if num == level then
         t = t - getCutsceneElapsed()
      end
      if num % 4 == 3 then
         t = t + WorldChangeTime
      elseif num % 4 == 2 then
         t = t + CutsceneTime3
      else
         t = t + CutsceneTime
      end
   end

   return t
end

function getCutsceneElapsed()
   if not cutsceneTimer then return 0 end
   return cutsceneTimer / FramesPerSecond
end


-- flags
FlagsAddr = 0xfa -- see 'startstate' below
--0xfb -- 224 (-32) during boss
--0xfc -- 16 during boss
--0xfd -- 16 during boss

function getFlags()
   local flags1 = memory.readbyte(FlagsAddr)
   local flags2 = memory.readbyte(FlagsAddr + 1)
   return flags1..','..flags2
end

function isCutsceneActive()
   -- sanity check
   if getLevelRemaining() ~= 0 then return false end
   
   -- for *-1, *-2, *-3
   if level % 4 ~= 3 then
      -- 0x40 changes from 0x81 to 0xc0 when the cutscene appears, and back when it disappears. this doesn't include the intro to the next level, but it doesn't affect the estimate, because CutsceneTime includes it. it only means the timer backs up and runs back into place at the beginning of these levels.
      if memory.readbyte(0x40) == 0xc0 then return true end
      
   -- for 1-4, 2-4, 3-4
   else
      -- 0x69 changes from 1 to 0 when the fortress catches on fire, and from 0 to 1 when you gain control after leaving the heli.
      if memory.readbyte(0x69) == 0 then return true end
   end

   return false
end

-- goes from 0 to F (for 1-1 to 4-4)
LevelIndexAddr = 0x049d

-- return 0--15 (meaning 1-1 to 4-4)
function getCurrentLevel()
   return memory.readbyte(LevelIndexAddr)
end

-- two bytes, least significant first
-- counts down as you move up
LevelPositionAddr = 0x7a
LevelStartPos = 0xfd00
LevelEndPos = 0xf200
LevelLength = LevelStartPos - LevelEndPos

-- return 1--0 (spawn to boss area)
function getLevelRemaining()
   local pos = memory.readword(LevelPositionAddr)
   -- this goes from 1 to 0 as you scroll the level to the boss area
   local r = (pos - LevelEndPos) / LevelLength
   -- this happens during the cutscene at the end of levels
   if r < 0 then r = 0 end
   return r
end


-- (linux only) return seconds since the epoch, with millisecond accuracy
function gettime()
   local f = assert(io.popen('date +%s%3N'))
   local t = assert(f:read('*a'))
   f:close()
   return tonumber(t) / 1000
end

function levelToString(n)
   return (math.floor(n / 4) + 1)..'-'..(n % 4 + 1)
end

-- convert frame count to 'MM:SS'
function formatTime(frames)
   local seconds = frames / FramesPerSecond
   local minutes = math.floor(seconds / 60)
   seconds = seconds % 60
   return string.format('%2d:%02d', minutes, seconds)
end


main()
